from plantcv.plantcv.hyperspectral.read_data import _find_closest
from plantcv.plantcv.hyperspectral.read_data import _make_pseudo_rgb
from plantcv.plantcv.hyperspectral.read_data import read_data
from plantcv.plantcv.hyperspectral.extract_wavelength import extract_wavelength
from plantcv.plantcv.hyperspectral.analyze_index import analyze_index
from plantcv.plantcv.hyperspectral.analyze_spectral import analyze_spectral
from plantcv.plantcv.hyperspectral.calibrate import calibrate
from plantcv.plantcv.hyperspectral._avg_reflectance import _avg_reflectance
from plantcv.plantcv.hyperspectral._inverse_covariance import _inverse_covariance

# add new functions to end of lists
__all__ = ["read_data", "_find_closest", "analyze_spectral", "analyze_index", "calibrate",
           "_make_pseudo_rgb", "extract_wavelength", "_avg_reflectance", "_inverse_covariance"]



# PlantCV classes
import os
import json
from plantcv.plantcv import fatal_error


class Params:
    """PlantCV parameters class."""

    def __init__(self, device=0, debug=None, debug_outdir=".", line_thickness=5, dpi=100, text_size=0.55,
                 text_thickness=2, marker_size=60, color_scale="gist_rainbow", color_sequence="sequential",
                 saved_color_scale=None, verbose=True):
        """Initialize parameters.

        Keyword arguments/parameters:
        device            = Device number. Used to count steps in the pipeline. (default: 0)
        debug             = None, print, or plot. Print = save to file, Plot = print to screen. (default: None)
        debug_outdir      = Debug images output directory. (default: .)
        line_thickness    = Width of line drawings. (default: 5)
        dpi               = Figure plotting resolution, dots per inch. (default: 100)
        text_size         = Size of plotting text. (default: 0.55)
        text_thickness    = Thickness of plotting text. (default: 2)
        marker_size       = Size of plotting markers (default: 60)
        color_scale       = Name of plotting color scale (matplotlib colormap). (default: gist_rainbow)
        color_sequence    = Build color scales in "sequential" or "random" order. (default: sequential)
        saved_color_scale = Saved color scale that will be applied next time color_palette is called. (default: None)
        verbose           = Whether or not in verbose mode. (default: True)

        :param device: int
        :param debug: str
        :param debug_outdir: str
        :param line_thickness: numeric
        :param dpi: int
        :param text_size: float
        :param text_thickness: int
        :param marker_size: int
        :param color_scale: str
        :param color_sequence: str
        :param saved_color_scale: list
        :param verbose: bool
        """
        self.device = device
        self.debug = debug
        self.debug_outdir = debug_outdir
        self.line_thickness = line_thickness
        self.dpi = dpi
        self.text_size = text_size
        self.text_thickness = text_thickness
        self.marker_size = marker_size
        self.color_scale = color_scale
        self.color_sequence = color_sequence
        self.saved_color_scale = saved_color_scale
        self.verbose = verbose


class Outputs:
    """PlantCV outputs class

    """

    def __init__(self):
        self.measurements = {}
        self.images = []
        self.observations = {}

        # Add a method to clear measurements
    def clear(self):
        self.measurements = {}
        self.images = []
        self.observations = {}

    # Method to add observation to outputs
    def add_observation(self, sample, variable, trait, method, scale, datatype, value, label):
        """
        Keyword arguments/parameters:
        sample       = Sample name. Used to distinguish between multiple samples
        variable     = A local unique identifier of a variable, e.g. a short name,
                       that is a key linking the definitions of variables with observations.
        trait        = A name of the trait mapped to an external ontology; if there is no exact mapping, an informative
                       description of the trait.
        method       = A name of the measurement method mapped to an external ontology; if there is no exact mapping, an
                       informative description of the measurement procedure
        scale        = Units of the measurement or scale in which the observations are expressed; if possible, standard
                       units and scales should be used and mapped to existing ontologies; in the case of non-standard
                       scale a full explanation should be given
        datatype     = The type of data to be stored, e.g. 'int', 'float', 'str', 'list', 'bool', etc.
        value        = The data itself
        label        = The label for each value (most useful when the data is a frequency table as in hue,
                       or other tables)

        :param sample: str
        :param variable: str
        :param trait: str
        :param method: str
        :param scale: str
        :param datatype: type
        :param value:
        :param label:
        """

        # Create an empty dictionary for the sample if it does not exist
        if sample not in self.observations:
            self.observations[sample] = {}

        # Supported data types
        supported_dtype = ["int", "float", "str", "list", "bool", "tuple", "dict", "NoneType", "numpy.float64"]
        # Supported class types
        class_list = [f"<class '{cls}'>" for cls in supported_dtype]

        # Send an error message if datatype is not supported by json
        if str(type(value)) not in class_list:
            # String list of supported types
            type_list = ', '.join(map(str, supported_dtype))
            fatal_error(f"The Data type {type(value)} is not compatible with JSON! Please use only these: {type_list}!")

        # Save the observation for the sample and variable
        self.observations[sample][variable] = {
            "trait": trait,
            "method": method,
            "scale": scale,
            "datatype": str(datatype),
            "value": value,
            "label": label
        }

    # Method to save observations to a file
    def save_results(self, filename, outformat="json"):
        """Save results to a file.

        Keyword arguments/parameters:
        filename       = Output filename
        outformat      = Output file format ("json" or "csv"). Default = "json"

        :param filename: str
        :param outformat: str
        """
        if outformat.upper() == "JSON":
            if os.path.isfile(filename):
                with open(filename, 'r') as f:
                    hierarchical_data = json.load(f)
                    hierarchical_data["observations"] = self.observations
            else:
                hierarchical_data = {"metadata": {}, "observations": self.observations}

            with open(filename, mode='w') as f:
                json.dump(hierarchical_data, f)
        elif outformat.upper() == "CSV":
            # Open output CSV file
            csv_table = open(filename, "w")
            # Write the header
            csv_table.write(",".join(map(str, ["sample", "trait", "value", "label"])) + "\n")
            # Iterate over data samples
            for sample in self.observations:
                # Iterate over traits for each sample
                for var in self.observations[sample]:
                    val = self.observations[sample][var]["value"]
                    # If the data type is a list or tuple we need to unpack the data
                    if isinstance(val, list) or isinstance(val, tuple):
                        # Combine each value with its label
                        for value, label in zip(self.observations[sample][var]["value"],
                                                self.observations[sample][var]["label"]):
                            # Skip list of tuple data types
                            if not isinstance(value, tuple):
                                # Save one row per value-label
                                row = [sample, var, value, label]
                                csv_table.write(",".join(map(str, row)) + "\n")
                    # If the data type is Boolean, store as a numeric 1/0 instead of True/False
                    elif isinstance(val, bool):
                        row = [sample,
                               var,
                               int(self.observations[sample][var]["value"]),
                               self.observations[sample][var]["label"]]
                        csv_table.write(",".join(map(str, row)) + "\n")
                    # For all other supported data types, save one row per trait
                    # Assumes no unusual data types are present (possibly a bad assumption)
                    else:
                        row = [sample,
                               var,
                               self.observations[sample][var]["value"],
                               self.observations[sample][var]["label"]
                               ]
                        csv_table.write(",".join(map(str, row)) + "\n")


class Spectral_data:
    # PlantCV Hyperspectral data class
    def __init__(self, array_data, max_wavelength, min_wavelength, max_value, min_value, d_type, wavelength_dict,
                 samples, lines, interleave, wavelength_units, array_type, pseudo_rgb, filename, default_bands):
        # The actual array/datacube
        self.array_data = array_data
        # Min/max available wavelengths (for spectral datacube)
        self.max_wavelength = max_wavelength
        self.min_wavelength = min_wavelength
        #  Min/max pixel value for single wavelength or index
        self.max_value = max_value
        self.min_value = min_value
        # Numpy data type
        self.d_type = d_type
        # Contains all available wavelengths where keys are wavelength and value are indices
        self.wavelength_dict = wavelength_dict
        # Resolution of a single band of spectral data is (samples, lines) rather than (x,y) with other arrays
        self.samples = samples
        self.lines = lines
        # Interleave type
        self.interleave = interleave
        self.wavelength_units = wavelength_units
        # The type of array data (entire datacube, specific index, first derivative, etc)
        self.array_type = array_type
        # Pseudo-RGB image if the array_type is a datacube
        self.pseudo_rgb = pseudo_rgb
        # The filename where the data originated from
        self.filename = filename
        # The default band indices needed to make an pseudo_rgb image, if not available then store None
        self.default_bands = default_bands

# Example
# spectral_array = Spectral_data(max_wavelength=1000.95, min_wavelength=379.027, d_type=numpy.float32,
#                           wavelength_dict=dictionary, samples=1600, lines=1704, interleave='bil',
#                           wavelength_units='nm', array_type="datacube", filename=fname, default_bands={159,253,520})



# Fuse two images

import os
import numpy as np
from skimage import img_as_ubyte
from plantcv.plantcv import fatal_error
from plantcv.plantcv import Spectral_data
from plantcv.plantcv import params
from plantcv.plantcv.hyperspectral.read_data import _make_pseudo_rgb
from plantcv.plantcv._debug import _debug


def image_fusion(img1, img2, wvs1, wvs2, array_type="multispectral"):
    """Fuse two images of the same size together to create a multispectral image
    img1: 1st image to be fused
    img2: 2nd image to be fused
    wvs1: list of wavelengths represent bands in img1
    wvs2: list of wavelengths represent bands in img2
    array_type: (optional) description of the fused array

    :param img1: numpy.ndarray
    :param img2: numpy.ndarray
    :param wvs1: list
    :param wvs2: list
    :param array_type: str
    :return fused_array: plantcv.Spectral_data
    """

    # If the image is 2D, expand to 3D to make stackable
    img1 = _expand_img_dims(img1)
    r1, c1, _ = img1.shape

    # If the image is 2D, expand to 3D to make stackable
    img2 = _expand_img_dims(img2)
    r2, c2, _ = img2.shape

    # Fatal error if images are not the same spatial dimensions
    if (r1, c1) != (r2, c2):
        fatal_error("Input images should have the same image size")

    # If the images are not the same data type, convert to 8-bit unsigned integer
    if img1.dtype != img2.dtype:
        img1 = img_as_ubyte(img1)
        img2 = img_as_ubyte(img2)

    # Concatenate the images on the depth/spectral (z) axis
    array_data = np.concatenate((img1, img2), axis=2)

    # sort all wavelengths
    wavelengths = np.array(wvs1 + wvs2)
    ind = np.argsort(wavelengths)
    wavelengths = wavelengths[ind]

    wavelength_dict = dict()
    for (idx, wv) in enumerate(wavelengths):
        wavelength_dict[wv] = float(idx)

    # sort array_data based on wavelengths
    array_data = array_data[:, :, ind]
    # Scale the array data to 0-1 by dividing by the maximum data type value
    array_data = (array_data / np.iinfo(array_data.dtype).max).astype(np.float32)

    r, c, b = array_data.shape

    fused_array = Spectral_data(array_data=array_data,
                                max_wavelength=float(max(wavelengths)),
                                min_wavelength=float(min(wavelengths)),
                                max_value=float(np.amax(array_data)),
                                min_value=float(np.amin(array_data)),
                                d_type=array_data.dtype,
                                wavelength_dict=wavelength_dict,
                                samples=c, lines=r, interleave="NA",
                                wavelength_units="nm", array_type=array_type,
                                pseudo_rgb=None, filename="NA", default_bands=None)

    # Make pseudo-rgb image and replace it inside the class instance object
    pseudo_rgb = _make_pseudo_rgb(fused_array)
    fused_array.pseudo_rgb = pseudo_rgb

    _debug(visual=pseudo_rgb, filename=os.path.join(params.debug_outdir, str(params.device) + "_fused_pseudo_rgb.png"))

    return fused_array


def _expand_img_dims(img):
    """Expand 2D images to 3D

    Inputs:
    img - input image

    Returns:
    img - image with expanded dimensions

    :params img: numpy.ndarray
    :return img: numpy.ndarray
    """
    # If the image is 2D, expand to 3D to make stackable
    if len(img.shape) == 2:
        return np.expand_dims(img, axis=2)
    # Return copy of image to break the reference to the input image
    return img.copy()



# RGB -> CMYK -> Gray

import cv2
import os
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import fatal_error
from plantcv.plantcv import params
import numpy as np


def rgb2gray_cmyk(rgb_img, channel):
    """Convert image from RGB colorspace to CMYK colorspace. Returns the specified subchannel as a gray image.

    Inputs:
    rgb_img   = RGB image data
    channel   = color subchannel (c = cyan, m = magenta, y = yellow, k=black)

    Returns:
    c | m | y | k = grayscale image from one CMYK color channel

    :param rgb_img: numpy.ndarray
    :param channel: str
    :return channel: numpy.ndarray
    """
    # The allowable channel inputs are c, m , y or k
    names = {"c": "cyan", "m": "magenta", "y": "yellow", "k": "black"}
    channel = channel.lower()
    if channel not in names:
        fatal_error("Channel " + str(channel) + " is not c, m, y or k!")

    # Create float
    bgr = rgb_img.astype(float)/255.

    # K channel
    k = 1 - np.max(bgr, axis=2)

    # C Channel
    c = (1 - bgr[..., 2] - k) / (1 - k)

    # M Channel
    m = (1 - bgr[..., 1] - k) / (1 - k)

    # Y Channel
    y = (1 - bgr[..., 0] - k) / (1 - k)

    # Convert the input BGR image to LAB colorspace
    cmyk = (np.dstack((c, m, y, k)) * 255).astype(np.uint8)
    # Split CMYK channels
    y, m, c, k = cv2.split(cmyk)
    # Create a channel dictionaries for lookups by a channel name index
    channels = {"c": c, "m": m, "y": y, "k": k}

    # Save or display the grayscale image
    _debug(visual=channels[channel], filename=os.path.join(params.debug_outdir,
                                                           str(params.device) + "_cmyk_" + names[channel] + ".png"))

    return channels[channel]



# Fill a mask using watershed and skeleton segments

import os
import cv2
import numpy as np
from skimage.segmentation import watershed
from plantcv.plantcv import outputs
from plantcv.plantcv import params
from plantcv.plantcv.visualize import colorize_label_img
from plantcv.plantcv._debug import _debug


def fill_segments(mask, objects, stem_objects=None, label="default"):
    """Fills masked segments from contours.

    Inputs:
    mask         = Binary image, single channel, object = 1 and background = 0
    objects      = List of contours

    Returns:
    filled_mask   = Labeled mask

    :param mask: numpy.ndarray
    :param objects: list
    :param stem_objects: numpy.ndarray
    :param label: str
    :return filled_mask: numpy.ndarray
    """

    h, w = mask.shape
    markers = np.zeros((h, w))

    objects_unique = objects.copy()
    if stem_objects is not None:
        objects_unique.append(np.vstack(stem_objects))

    labels = np.arange(len(objects_unique)) + 1
    for i, l in enumerate(labels):
        cv2.drawContours(markers, objects_unique, i, int(l), 5)

    # Fill as a watershed segmentation from contours as markers
    filled_mask = watershed(mask == 0, markers=markers,
                            mask=mask != 0, compactness=0)

    # Count area in pixels of each segment
    ids, counts = np.unique(filled_mask, return_counts=True)

    if stem_objects is None:
        outputs.add_observation(sample=label, variable='segment_area', trait='segment area',
                                method='plantcv.plantcv.morphology.fill_segments',
                                scale='pixels', datatype=list,
                                value=counts[1:].tolist(),
                                label=(ids[1:]-1).tolist())
    else:
        outputs.add_observation(sample=label, variable='leaf_area', trait='segment area',
                                method='plantcv.plantcv.morphology.fill_segments',
                                scale='pixels', datatype=list,
                                value=counts[1:-1].tolist(),
                                label=(ids[1:-1]-1).tolist())
        outputs.add_observation(sample=label, variable='stem_area', trait='segment area',
                                method='plantcv.plantcv.morphology.fill_segments',
                                scale='pixels', datatype=list,
                                value=counts[-1].tolist(),
                                label=(ids[-1]-1).tolist())

    # rgb_vals = color_palette(num=len(labels), saved=False)
    # filled_img = np.zeros((h, w, 3), dtype=np.uint8)
    # for l in labels:
    #     for ch in range(3):
    #         filled_img[:, :, ch][filled_mask == l] = rgb_vals[l - 1][ch]

    debug = params.debug
    params.debug = None
    filled_img = colorize_label_img(filled_mask)
    params.debug = debug
    _debug(visual=filled_img, filename=os.path.join(params.debug_outdir,
                                                    str(params.device) + "_filled_segments_img.png"))

    return filled_mask



# Analyze stem characteristics

import os
import cv2
import numpy as np
from plantcv.plantcv import params
from plantcv.plantcv import outputs
from plantcv.plantcv import plot_image
from plantcv.plantcv import print_image


def analyze_stem(rgb_img, stem_objects, label="default"):
    """ Calculate angle of segments (in degrees) by fitting a linear regression line to segments.

        Inputs:
        rgb_img       = RGB image to plot debug image
        stem_objects  = List of stem segments (output from segment_sort function)
        label        = optional label parameter, modifies the variable name of observations recorded

        Returns:
        labeled_img    = Stem analysis debugging image


        :param rgb_img: numpy.ndarray
        :param stem_objects: list
        :param label: str
        :return labeled_img: numpy.ndarray
    """
    params.device += 1
    labeled_img = np.copy(rgb_img)
    img_x, img_y, _ = np.shape(labeled_img)
    grouped_stem = np.vstack(stem_objects)

    # Find vertical height of the stem by measuring bounding box
    stem_x, stem_y, width, height = cv2.boundingRect(grouped_stem)

    # Calculate stem angle
    [vx, vy, x, y] = cv2.fitLine(grouped_stem, cv2.DIST_L2, 0, 0.01, 0.01)
    slope = -vy / vx

    # Calculate stem path length
    stem_length = cv2.arcLength(grouped_stem, False) / 2

    outputs.add_observation(sample=label, variable='stem_height', trait='vertical length of stem segments',
                            method='plantcv.plantcv.morphology.analyze_stem', scale='pixels', datatype=float,
                            value=height, label=None)
    outputs.add_observation(sample=label, variable='stem_angle', trait='angle of combined stem object',
                            method='plantcv.plantcv.morphology.analyze_stem', scale='degrees', datatype=float,
                            value=float(slope), label=None)
    outputs.add_observation(sample=label, variable='stem_length', trait='path length of combined stem object',
                            method='plantcv.plantcv.morphology.analyze_stem', scale='None', datatype=float,
                            value=stem_length, label=None)

    if params.debug is not None:
        # Draw culm_height
        cv2.line(labeled_img, (int(stem_x), stem_y), (int(stem_x), stem_y + height), (0, 255, 0), params.line_thickness)
        # Draw combined stem angle
        x_min = 0  # Set bounds for regression lines to get drawn
        x_max = img_x
        intercept1 = int(((x - x_min) * slope) + y)
        intercept2 = int(((x - x_max) * slope) + y)
        if slope > 1000000 or slope < -1000000:
            print("Slope  is ", slope, " and cannot be plotted.")
        else:
            cv2.line(labeled_img, (x_max - 1, intercept2), (x_min, intercept1), (0, 0, 255), 1)
        if params.debug == 'print':
            print_image(labeled_img, os.path.join(params.debug_outdir, str(params.device) + 'stem_analze.png'))
        elif params.debug == 'plot':
            plot_image(labeled_img)

    return labeled_img



# Debugging module

from plantcv.plantcv import params
from plantcv.plantcv import print_image
from plantcv.plantcv import plot_image


def _debug(visual, filename=None, **kwargs):
    """Save or display a visual for debugging.

    Inputs:
    visual   - An image or plot to display for debugging
    filename - An optional filename to save the visual to (default: None)

    :param visual: numpy.ndarray
    :param filename: str
    """
    # Auto-increment the device counter
    params.device += 1
    
    if params.debug == "print":
        # If debug is print, save the image to a file
        print_image(img=visual, filename=filename)
    elif params.debug == "plot":
        # If debug is plot, print to the plotting device
        plot_image(img=visual, **kwargs)



# Analyze reflectance signal data in an index

import os
import numpy as np
from plantcv.plantcv import params
from plantcv.plantcv import outputs
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import fatal_error
from plotnine import labs
from plantcv.plantcv.visualize import histogram
from plantcv.plantcv import deprecation_warning


def analyze_index(index_array, mask, bins=100, min_bin=0, max_bin=1, histplot=None, label="default"):
    """This extracts the hyperspectral index statistics and writes the values  as observations out to
       the Outputs class.

    Inputs:
    index_array  = Instance of the Spectral_data class, usually the output from pcv.hyperspectral.extract_index
    mask         = Binary mask made from selected contours
    histplot     = if True plots histogram of intensity values
    bins         = optional, number of classes to divide spectrum into
    min_bin      = optional, minimum bin value ("auto" or user input minimum value)
    max_bin      = optional, maximum bin value ("auto" or user input maximum value)
    label        = optional label parameter, modifies the variable name of observations recorded



    :param index_array: __main__.Spectral_data
    :param mask: numpy array
    :param histplot: bool
    :param bins: int
    :param max_bin: float, str
    :param min_bin: float, str
    :param label: str
    :return analysis_image: ggplot, None
    """
    if histplot is not None:
        deprecation_warning("'histplot' will be deprecated in a future version of PlantCV. "
                            "This function creates a histogram by default.")

    debug = params.debug
    params.debug = None

    if len(np.shape(mask)) > 2 or len(np.unique(mask)) > 2:
        fatal_error("Mask should be a binary image of 0 and nonzero values.")

    if len(np.shape(index_array.array_data)) > 2:
        fatal_error("index_array data should be a grayscale image.")

    # Mask data and collect statistics about pixels within the masked image
    masked_array = index_array.array_data[np.where(mask > 0)]
    masked_array = masked_array[np.isfinite(masked_array)]

    index_mean = np.nanmean(masked_array)
    index_median = np.nanmedian(masked_array)
    index_std = np.nanstd(masked_array)

    # Set starting point and max bin values
    maxval = max_bin
    b = min_bin

    # Calculate observed min and max pixel values of the masked array
    observed_max = np.nanmax(masked_array)
    observed_min = np.nanmin(masked_array)

    # Auto calculate max_bin if set
    if type(max_bin) is str and (max_bin.upper() == "AUTO"):
        maxval = float(round(observed_max, 8))  # Auto bins will detect maxval to use for calculating labels/bins
    if type(min_bin) is str and (min_bin.upper() == "AUTO"):
        b = float(round(observed_min, 8))  # If bin_min is auto then overwrite starting value

    # Print a warning if observed min/max outside user defined range
    if observed_max > maxval or observed_min < b:
        print("WARNING!!! The observed range of pixel values in your masked index provided is [" + str(observed_min) +
              ", " + str(observed_max) + "] but the user defined range of bins for pixel frequencies is [" + str(b) +
              ", " + str(maxval) + "]. Adjust min_bin and max_bin in order to avoid cutting off data being collected.")

    # Calculate histogram
    hist_fig, hist_data = histogram(index_array.array_data, mask=mask, bins=bins, lower_bound=b, upper_bound=maxval,
                                    hist_data=True)
    bin_labels, hist_percent = hist_data['pixel intensity'].tolist(), hist_data['proportion of pixels (%)'].tolist()

    # Restore user debug setting
    params.debug = debug
    hist_fig = hist_fig + labs(x='Index Reflectance', y='Proportion of pixels (%)')

    # Print or plot histogram
    _debug(visual=hist_fig,
           filename=os.path.join(params.debug_outdir, str(params.device) + index_array.array_type + "_hist.png"))

    analysis_image = hist_fig

    outputs.add_observation(sample=label, variable='mean_' + index_array.array_type,
                            trait='Average ' + index_array.array_type + ' reflectance',
                            method='plantcv.plantcv.hyperspectral.analyze_index', scale='reflectance', datatype=float,
                            value=float(index_mean), label='none')

    outputs.add_observation(sample=label, variable='med_' + index_array.array_type,
                            trait='Median ' + index_array.array_type + ' reflectance',
                            method='plantcv.plantcv.hyperspectral.analyze_index', scale='reflectance', datatype=float,
                            value=float(index_median), label='none')

    outputs.add_observation(sample=label, variable='std_' + index_array.array_type,
                            trait='Standard deviation ' + index_array.array_type + ' reflectance',
                            method='plantcv.plantcv.hyperspectral.analyze_index', scale='reflectance', datatype=float,
                            value=float(index_std), label='none')

    outputs.add_observation(sample=label, variable='index_frequencies_' + index_array.array_type,
                            trait='index frequencies', method='plantcv.plantcv.analyze_index', scale='frequency',
                            datatype=list, value=hist_percent, label=bin_labels)

    # Print or plot the masked image
    _debug(visual=masked_array,
           filename=os.path.join(params.debug_outdir, str(params.device) + index_array.array_type + ".png"))
    # Store images
    outputs.images.append(analysis_image)

    return analysis_image



from plantcv.plantcv.transform import rotate as rotate_new
import sys


def rotate(img, rotation_deg, crop):
    """Rotate image, sometimes it is necessary to rotate image, especially when clustering for
       multiple plants is needed.

    Inputs:
    img          = RGB or grayscale image data
    rotation_deg = rotation angle in degrees, can be a negative number,
                   positive values move counter clockwise.
    crop         = either true or false, if true, dimensions of rotated image will be same as original image.

    Returns:
    rotated_img  = rotated image

    :param img: numpy.ndarray
    :param rotation_deg: double
    :param crop: bool
    :return rotated_img: numpy.ndarray
    """
    print("""Deprecation Warning:
    plantcv.rotate has moved to plantcv.transform.rotate. 
    plantcv.rotate will be removed in a future version""", file=sys.stderr)
    rotated_img = rotate_new(img, rotation_deg, crop)
    return rotated_img



# Analyze reflectance signal hyperspectral images

import os
import numpy as np
import pandas as pd
from plantcv.plantcv import params
from plantcv.plantcv import outputs
from plotnine import ggplot, aes, geom_line, scale_x_continuous
from plantcv.plantcv import deprecation_warning
from plantcv.plantcv._debug import _debug


def analyze_spectral(array, mask, histplot=None, label="default"):
    """This extracts the hyperspectral reflectance values of each pixel writes the values out to
       a file. It can also print out a histogram plot of pixel intensity
       and a pseudocolor image of the plant.

    Inputs:
    array        = Hyperspectral data instance
    mask         = Binary mask made from selected contours
    histplot     = (to be deprecated) if True plots histogram of reflectance intensity values
    label        = optional label parameter, modifies the variable name of observations recorded

    Returns:
    analysis_img = output image

    :param array: __main__.Spectral_data
    :param mask: numpy array
    :param histplot: bool
    :param label: str
    :return analysis_img: ggplot
    """
    if histplot is not None:
        deprecation_warning("'histplot' will be deprecated in a future version of PlantCV. "
                            "Instead of a histogram this function plots the mean of spectra in the masked area.")

    array_data = array.array_data

    # List of wavelengths recorded created from parsing the header file will be string, make list of floats
    wavelength_data = array_data[np.where(mask > 0)]

    # Calculate mean reflectance across wavelengths
    wavelength_freq = wavelength_data.mean(axis=0)
    max_per_band = wavelength_data.max(axis=0)
    min_per_band = wavelength_data.min(axis=0)
    std_per_band = wavelength_data.std(axis=0)

    # Identify smallest and largest wavelengths available to scale the x-axis
    min_wavelength = array.min_wavelength
    max_wavelength = array.max_wavelength

    # Create lists with wavelengths in float format rather than as strings
    # and make a list of the frequencies since they are in an array
    new_wavelengths = []
    new_freq = []
    new_std_per_band = []
    new_max_per_band = []
    new_min_per_band = []

    for i, wavelength in enumerate(array.wavelength_dict):
        new_wavelengths.append(wavelength)
        new_freq.append((wavelength_freq[i]).astype(float))
        new_std_per_band.append(std_per_band[i].astype(float))
        new_max_per_band.append(max_per_band[i].astype(float))
        new_min_per_band.append(min_per_band[i].astype(float))

    # Calculate reflectance statistics
    avg_reflectance = np.average(wavelength_data)
    std_reflectance = np.std(wavelength_data)
    median_reflectance = np.median(wavelength_data)

    wavelength_labels = []
    for i in array.wavelength_dict.keys():
        wavelength_labels.append(i)

    # Store data into outputs class
    outputs.add_observation(sample=label, variable='global_mean_reflectance', trait='global mean reflectance',
                            method='plantcv.plantcv.hyperspectral.analyze_spectral', scale='reflectance',
                            datatype=float, value=float(avg_reflectance), label='reflectance')
    outputs.add_observation(sample=label, variable='global_median_reflectance', trait='global median reflectance',
                            method='plantcv.plantcv.hyperspectral.analyze_spectral', scale='reflectance',
                            datatype=float, value=float(median_reflectance), label='reflectance')
    outputs.add_observation(sample=label, variable='global_spectral_std',
                            trait='pixel-wise standard deviation per band',
                            method='plantcv.plantcv.hyperspectral.analyze_spectral', scale='None', datatype=float,
                            value=float(std_reflectance), label='reflectance')
    outputs.add_observation(sample=label, variable='global_spectral_std', trait='pixel-wise standard deviation ',
                            method='plantcv.plantcv.hyperspectral.analyze_spectral', scale='None', datatype=float,
                            value=float(std_reflectance), label='reflectance')
    outputs.add_observation(sample=label, variable='max_reflectance', trait='maximum reflectance per band',
                            method='plantcv.plantcv.hyperspectral.analyze_spectral', scale='reflectance', datatype=list,
                            value=new_max_per_band, label=wavelength_labels)
    outputs.add_observation(sample=label, variable='min_reflectance', trait='minimum reflectance per band',
                            method='plantcv.plantcv.hyperspectral.analyze_spectral', scale='reflectance', datatype=list,
                            value=new_min_per_band, label=wavelength_labels)
    outputs.add_observation(sample=label, variable='spectral_std', trait='pixel-wise standard deviation per band',
                            method='plantcv.plantcv.hyperspectral.analyze_spectral', scale='None', datatype=list,
                            value=new_std_per_band, label=wavelength_labels)
    outputs.add_observation(sample=label, variable='spectral_frequencies', trait='spectral frequencies',
                            method='plantcv.plantcv.hyperspectral.analyze_spectral', scale='frequency', datatype=list,
                            value=new_freq, label=wavelength_labels)

    dataset = pd.DataFrame({'Wavelength (' + array.wavelength_units + ')': new_wavelengths,
                            'Reflectance': wavelength_freq})
    mean_spectra = (ggplot(data=dataset,
                    mapping=aes(x='Wavelength (' + array.wavelength_units + ')', y='Reflectance'))
                    + geom_line(color='purple')
                    + scale_x_continuous(breaks=list(range(int(np.floor(min_wavelength)),
                                                           int(np.ceil(max_wavelength)), 50)))
                    )

    analysis_img = mean_spectra

    _debug(visual=mean_spectra, filename=os.path.join(params.debug_outdir, str(params.device) + "_mean_spectra.png"))

    return analysis_img



# Crop

import os
import cv2
import numpy as np
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import params


def crop(img, x, y, h, w):
    """
    Crop image.

       Inputs:
       img       = RGB, grayscale, or hyperspectral image data
       x         = X coordinate of starting point
       y         = Y coordinate of starting point
       h         = Height
       w         = Width

       Returns:
       cropped   = cropped image

       :param img: numpy.ndarray
       :param x: int
       :param y: int
       :param h: int
       :param w: int
       :return cropped: numpy.ndarray
       """

    # Check if the array data format
    if len(np.shape(img)) > 2 and np.shape(img)[-1] > 3:
        ref_img = img[:, :, [0]]
        ref_img = np.transpose(np.transpose(ref_img)[0])
        cropped = img[y:y + h, x:x + w, :]
    else:
        ref_img = np.copy(img)
        cropped = img[y:y + h, x:x + w]

    # Create the rectangle contour vertices
    pt1 = (x, y)
    pt2 = (x + w - 1, y + h - 1)

    ref_img = cv2.rectangle(img=ref_img, pt1=pt1, pt2=pt2, color=(255, 0, 0), thickness=params.line_thickness)

    _debug(visual=ref_img,
           filename=os.path.join(params.debug_outdir, str(params.device) + "_crop.png"))

    return cropped



from plantcv.plantcv.transform.color_correction import get_color_matrix
from plantcv.plantcv.transform.color_correction import get_matrix_m
from plantcv.plantcv.transform.color_correction import calc_transformation_matrix
from plantcv.plantcv.transform.color_correction import apply_transformation_matrix
from plantcv.plantcv.transform.color_correction import save_matrix
from plantcv.plantcv.transform.color_correction import load_matrix
from plantcv.plantcv.transform.color_correction import correct_color
from plantcv.plantcv.transform.color_correction import create_color_card_mask
from plantcv.plantcv.transform.color_correction import quick_color_check
from plantcv.plantcv.transform.color_correction import find_color_card
from plantcv.plantcv.transform.rescale import rescale
from plantcv.plantcv.transform.rotate import rotate
from plantcv.plantcv.transform.nonuniform_illumination import nonuniform_illumination
from plantcv.plantcv.transform.resize import resize, resize_factor
from plantcv.plantcv.transform.warp import warp, warp_align
from plantcv.plantcv.transform.gamma_correct import gamma_correct

__all__ = ["get_color_matrix", "get_matrix_m", "calc_transformation_matrix", "apply_transformation_matrix",
           "save_matrix", "load_matrix", "correct_color", "create_color_card_mask", "quick_color_check",
           "find_color_card", "rescale", "nonuniform_illumination", "resize", "resize_factor",
           "warp", "rotate", "warp", "warp_align", "gamma_correct"]



import os
import json


def json2csv(json_file, csv_file):
    if os.path.exists(json_file):
        # If the JSON file exists open it for reading
        json_data = open(json_file, "r")
        data = json.load(json_data)
        # If the data is JSON but it does not have the components we expect from PlantCV raise an error
        if "variables" not in data or "entities" not in data:
            raise ValueError("Invalid JSON file: {0}".format(json_file))

        # Split up variables
        meta_vars = []
        scalar_vars = []
        multi_vars = []
        for key, var in data["variables"].items():
            if var["category"] == "metadata":
                meta_vars.append(key)
            elif var["datatype"] in ["<class 'bool'>", "<class 'int'>", "<class 'float'>", "<class 'str'>",
                                     "<type 'bool'>", "<type 'int'>", "<type 'float'>", "<type 'str'>"]:
                scalar_vars.append(key)
            elif var["datatype"] in ["<class 'list'>", "<type 'list'>"]:
                multi_vars.append(key)

        # Create a CSV file of single-value traits
        csv = open(csv_file + "-single-value-traits.csv", "w")

        # Build the single-value variables output table
        csv.write(",".join(map(str, meta_vars + ["sample"] + scalar_vars)) + "\n")
        for entity in data["entities"]:
            row = []
            # Add metadata variables
            for var in meta_vars:
                obs = entity[data["variables"][var]["category"]]
                if var in obs:
                    row.append(obs[var]["value"])
                else:
                    row.append("NA")
            # Add scalar variables
            for sample in entity["observations"]:
                measurements = [sample]
                for var in scalar_vars:
                    obs = entity[data["variables"][var]["category"]][sample]
                    if var in obs:
                        measurements.append(obs[var]["value"])
                    else:
                        measurements.append("NA")
                csv.write(",".join(map(str, row + measurements)) + "\n")
        # Close the CSV file
        csv.close()

        # Create a CSV file of multi-value variables
        csv = open(csv_file + "-multi-value-traits.csv", "w")
        csv.write(",".join(map(str, meta_vars + ["sample", "trait", "value", "label"])) + "\n")
        for entity in data["entities"]:
            meta_row = []
            # Add metadata variables
            for var in meta_vars:
                obs = entity[data["variables"][var]["category"]]
                if var in obs:
                    meta_row.append(obs[var]["value"])
                else:
                    meta_row.append("NA")
            # Add multi-value variables
            for sample in entity["observations"]:
                for var in multi_vars:
                    obs = entity[data["variables"][var]["category"]][sample]
                    if var in obs:
                        if obs[var]["label"] != "none":
                            for i in range(0, len(obs[var]["value"])):
                                row = [sample, var, obs[var]["value"][i], obs[var]["label"][i]]
                                csv.write(",".join(map(str, meta_row + row)) + "\n")
                    else:
                        csv.write(",".join(map(str, meta_row + [var, "NA", "NA"])) + "\n")
        csv.close()
    else:
        # If the file does not exist raise an error
        raise IOError("File does not exist: {0}".format(json_file))


def tabulate_bayes_classes(input_file, output_file):
    """Tabulate pixel RGB values into a table for naive Bayes training.
    Inputs:
    input_file   = Input text file of class names and RGB values
    output_file  = Output file for storing the tab-delimited naive Bayes training data

    The input file should have class names preceded by the "#" character. RGB values can be pasted
    directly from ImageJ without reformatting. E.g.:

    #plant
    96,154,72	95,153,72	91,155,71	91,160,70	90,155,67	92,152,66	92,157,70
    54,104,39	56,104,38	59,106,41	57,105,43	54,104,40	54,103,35	56,101,39	58,99,41	59,99,41
    #background
    114,127,121	117,135,125	120,137,131	132,145,138	142,154,148	151,166,158	160,182,172
    115,125,121	118,131,123	122,132,135	133,142,144	141,151,152	150,166,158	159,179,172

    :param input_file: str
    :param output_file: str
    """

    # If the input file does not exist raise an error
    if not os.path.exists(input_file):
        raise IOError("File does not exist: {0}".format(input_file))

    # Read the file into a string
    with open(input_file, "r") as fd:
        pixel_data = fd.read()

    # Split the file string by the classname/header character
    classes = pixel_data.split("#")

    # Parse the class data
    class_dict = {}
    # Ignore the first item, it's the empty string or whitespace to the "left" of the first class/header
    for i in range(1, len(classes)):
        # Replace tabs with newlines
        classes[i] = classes[i].replace("\t", "\n")
        # Split the class data on newlines
        class_data = classes[i].split("\n")
        # The class name is the first item
        class_name = class_data[0]
        rgb_values = []
        # Loop over the RGB values, starting with the second element
        for j in range(1, len(class_data)):
            # Skip blank lines but keep the lines with RGB values
            if len(class_data[j]) > 0:
                rgb_values.append(class_data[j])
        class_dict[class_name] = rgb_values

    # Each class could have a different number of RGB values, find the largest
    total_rgb = 0
    for class_name in class_dict:
        if len(class_dict[class_name]) > total_rgb:
            total_rgb = len(class_dict[class_name])

    # Pad the classes with empty strings if they have less than the total RGB values
    for class_name in class_dict:
        missing = total_rgb - len(class_dict[class_name])
        if missing > 0:
            for i in range(missing):
                class_dict[class_name].append("")

    # Open the output file
    out = open(output_file, "w")
    # Create the output table
    class_names = class_dict.keys()
    out.write("\t".join(map(str, class_names)) + "\n")
    for i in range(0, total_rgb):
        row = []
        for class_name in class_names:
            row.append(class_dict[class_name][i])
        out.write("\t".join(map(str, row)) + "\n")
    out.close()



# RGB -> Gray

import cv2
import os
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug


def rgb2gray(rgb_img):
    """Convert image from RGB colorspace to Gray.

    Inputs:
    rgb_img    = RGB image data

    Returns:
    gray   = grayscale image

    :param rgb_img: numpy.ndarray
    :return gray: numpy.ndarray
    """

    gray = cv2.cvtColor(rgb_img, cv2.COLOR_BGR2GRAY)

    _debug(visual=gray, filename=os.path.join(params.debug_outdir, str(params.device) + "_gray.png"))

    return gray



from plantcv.plantcv.spectral_index.spectral_index import ndvi
from plantcv.plantcv.spectral_index.spectral_index import gdvi
from plantcv.plantcv.spectral_index.spectral_index import savi
from plantcv.plantcv.spectral_index.spectral_index import pri
from plantcv.plantcv.spectral_index.spectral_index import ari
from plantcv.plantcv.spectral_index.spectral_index import ci_rededge
from plantcv.plantcv.spectral_index.spectral_index import cri550
from plantcv.plantcv.spectral_index.spectral_index import cri700
from plantcv.plantcv.spectral_index.spectral_index import egi
from plantcv.plantcv.spectral_index.spectral_index import evi
from plantcv.plantcv.spectral_index.spectral_index import mari
from plantcv.plantcv.spectral_index.spectral_index import mcari
from plantcv.plantcv.spectral_index.spectral_index import mtci
from plantcv.plantcv.spectral_index.spectral_index import ndre
from plantcv.plantcv.spectral_index.spectral_index import psnd_chla
from plantcv.plantcv.spectral_index.spectral_index import psnd_chlb
from plantcv.plantcv.spectral_index.spectral_index import psnd_car
from plantcv.plantcv.spectral_index.spectral_index import psri
from plantcv.plantcv.spectral_index.spectral_index import pssr_chla
from plantcv.plantcv.spectral_index.spectral_index import pssr_chlb
from plantcv.plantcv.spectral_index.spectral_index import pssr_car
from plantcv.plantcv.spectral_index.spectral_index import rgri
from plantcv.plantcv.spectral_index.spectral_index import rvsi
from plantcv.plantcv.spectral_index.spectral_index import sipi
from plantcv.plantcv.spectral_index.spectral_index import sr
from plantcv.plantcv.spectral_index.spectral_index import vari
from plantcv.plantcv.spectral_index.spectral_index import vi_green
from plantcv.plantcv.spectral_index.spectral_index import wi


# add new functions to end of lists
__all__ = ["ndvi", "gdvi", "savi", "pri", "ari", "ci_rededge", "cri550", "cri700", "egi", "evi", "mari", "mcari",
           "mtci", "ndre", "psnd_chla", "psnd_chlb", "psri", "pssr_chla", "pssr_chlb", "pssr_car", "rgri", "rvsi",
           "sipi", "sr", "vi_green", "wi"]



# Join images (XOR)

import cv2
import os
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import params


def logical_xor(bin_img1, bin_img2):
    """
    Join two images using the bitwise XOR operator.

    Inputs:
    bin_img1   = Binary image data to be compared to bin_img2
    bin_img2   = Binary image data to be compared to bin_img1

    Returns:
    merged     = joined binary image

    :param bin_img1: numpy.ndarray
    :param bin_img2: numpy.ndarray
    :return merged: numpy.ndarray
    """

    merged = cv2.bitwise_xor(bin_img1, bin_img2)

    _debug(visual=merged,
           filename=os.path.join(params.debug_outdir, str(params.device) + '_xor_joined.png'),
           cmap='gray')

    return merged



# Compare auto threshold methods for a grayscale image

import os
import cv2
import numpy as np
from plantcv.plantcv import params
from plantcv.plantcv.transform import resize_factor
from plantcv.plantcv import fatal_error
from plantcv.plantcv._debug import _debug
from plantcv.plantcv.threshold import mean
from plantcv.plantcv.threshold import otsu
from plantcv.plantcv.threshold import gaussian
from plantcv.plantcv.threshold import triangle


def auto_threshold_methods(gray_img, grid_img=True, object_type="light"):
    """ Compare auto threshold methods for a grayscale image

    Inputs:
    gray_img     = Grayscale image data
    grid_img     = Whether or not to compile masks into a single plot
    object_type  = "light" or "dark" (default: "light")
                   - If object is lighter than the background then standard thresholding is done
                   - If object is darker than the background then inverse thresholding is done

    Returns:
    labeled_imgs = List of labeled plotting images

    :param gray_img: numpy.ndarray
    :param grid_img: bool
    :param object_type: str
    :return labeled_imgs: list

    """
    # Check that the image is grayscale
    if not len(np.shape(gray_img)) == 2:
        fatal_error("Input image is not grayscale!")

    # Store and disable debug mode
    debug = params.debug
    params.debug = None

    # Initialize threshold method names, mask list, final images
    method_names = ["Gaussian", "Mean", "Otsu", "Triangle"]
    all_methods = []
    labeled_imgs = []

    # Set starting location for labeling different masks
    y = int(np.shape(gray_img)[0] / 8)
    x = int(np.shape(gray_img)[1] / 8)

    # Create mask imgs from each thresholding method
    all_methods.append(gaussian(gray_img=gray_img, max_value=255, object_type=object_type))
    all_methods.append(mean(gray_img=gray_img, max_value=255, object_type=object_type))
    all_methods.append(otsu(gray_img=gray_img, max_value=255, object_type=object_type))
    all_methods.append(triangle(gray_img=gray_img, max_value=255, object_type=object_type, xstep=1))

    # Plot labels of each colorspace on the corresponding img
    for i, method in enumerate(all_methods):
        converted_img = cv2.cvtColor(method, cv2.COLOR_GRAY2RGB)
        labeled = cv2.putText(img=converted_img, text=method_names[i], org=(x, y),
                              fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                              fontScale=params.text_size, color=(255, 0, 255), thickness=params.text_thickness)
        # Reset debug mode
        params.debug = debug
        _debug(visual=labeled,
               filename=os.path.join(params.debug_outdir,
                                     str(params.device) + "_" + method_names[i] + "_vis_thresholds.png"))
        labeled_imgs.append(labeled)

    if grid_img:
        # Store and disable debug mode
        debug = params.debug
        params.debug = None
        # Compile images together into one
        top_row = np.hstack([labeled_imgs[0], labeled_imgs[1]])
        bot_row = np.hstack([labeled_imgs[2], labeled_imgs[3]])
        plotting_img = np.vstack([top_row, bot_row])
        labeled_imgs.append(plotting_img)
        plotting_img = resize_factor(plotting_img, factors=(0.5, 0.5))
        # Reset debug mode
        params.debug = debug
        _debug(visual=plotting_img,
               filename=os.path.join(params.debug_outdir, str(params.device) + "_vis_all_thresholds.png"))

    return labeled_imgs



# Print Numerical Data

from plantcv.plantcv import outputs


def print_results(filename):
    """Print result table

    Inputs:
    filename = filename

    :param filename: str
    :return:
    """
    print("""Deprecation warning: plantcv.print_results will be removed in a future version.
             Please use plantcv.outputs.save_results instead.
          """)
    outputs.save_results(filename=filename, outformat="json")



# Deprecation handling

import sys
from plantcv.plantcv import _version
from plantcv.plantcv import params

def deprecation_warning(warning):
    """Print out deprecation warning

    Inputs:
    warning = warning message text

    :param warning: str
    """
    version = _version.get_versions()
    warning_msg = f"DeprecationWarning: {warning} Current PlantCV version: {version['version']} released on {version['date']}"
    if params.verbose is True:
        print(warning_msg, file=sys.stderr)



# Script to identify corners/acute angles of an object

import os
import cv2
import numpy as np
import math
from plantcv.plantcv import params
from plantcv.plantcv import outputs
from plantcv.plantcv._debug import _debug


def acute_vertex(img, obj, win, thresh, sep, label="default"):
    """acute_vertex: identify corners/acute angles of an object

    For each point in contour, get a point before (pre) and after (post) the point of interest,
    calculate the angle between the pre and post point.

    Inputs:
    img    = the original image
    obj    = a contour of the plant object (this should be output from the object_composition.py fxn)
    win    = win argument specifies the pre and post point distances (a value of 30 worked well for a sample image)
    thresh = an threshold to set for acuteness; keep points with an angle more acute than the threshold (a value of 15
             worked well for sample image)
    sep    = the number of contour points to search within for the most acute value
    label  = optional label parameter, modifies the variable name of observations recorded


    Returns:
    acute_points = list of acute points
    img2         = debugging image

    :param img: ndarray
    :param obj: ndarray
    :param win: int
    :param thresh: int
    :param sep: int
    :param label: str
    :return acute_points: ndarray
    :return img2: ndarray
    """
    params.device += 1
    chain = []
    if not np.any(obj):
        acute = ('NA', 'NA')
        return acute
    for i in range(len(obj) - win):
        x, y = obj[i].ravel()
        pre_x, pre_y = obj[i - win].ravel()
        post_x, post_y = obj[i + win].ravel()

        # Angle in radians derived from Law of Cosines, converted to degrees
        P12 = np.sqrt((x - pre_x) * (x - pre_x) + (y - pre_y) * (y - pre_y))
        P13 = np.sqrt((x - post_x) * (x - post_x) + (y - post_y) *
                      (y - post_y))
        P23 = np.sqrt((pre_x - post_x) * (pre_x - post_x) + (pre_y - post_y) *
                      (pre_y - post_y))
        if (2 * P12 * P13) > 0.001:
            dot = (P12 * P12 + P13 * P13 - P23 * P23) / (2 * P12 * P13)
        elif (2 * P12 * P13) < 0.001:
            dot = (P12 * P12 + P13 * P13 - P23 * P23) / 0.001

        if dot < -1:  # If float exceeds -1 prevent arcos error and force to equal -1
            dot = -1
        ang = math.degrees(math.acos(dot))
        chain.append(ang)

    # Select points in contour that have an angle more acute than thresh
    index = []
    for c in range(len(chain)):
        if float(chain[c]) <= thresh:
            index.append(c)
    # There oftentimes several points around tips with acute angles
    # Here we try to pick the most acute angle given a set of contiguous point
    # Sep is the number of points to evaluate the number of vertices
    out = []
    tester = []
    for i in range(len(index) - 1):
        # print str(index[i])
        if index[i + 1] - index[i] < sep:
            tester.append(index[i])
        if index[i + 1] - index[i] >= sep:
            tester.append(index[i])
            # print(tester)
            angles = ([chain[d] for d in tester])
            keeper = angles.index(min(angles))
            t = tester[keeper]
            # print str(t)
            out.append(t)
            tester = []

    # Store the points in the variable acute
    acute = obj[[out]]
    acute_points = []
    for pt in acute:
        acute_points.append(pt[0].tolist())

    img2 = np.copy(img)
    # Plot each of these tip points on the image
    for i in acute:
        x, y = i.ravel()
        cv2.circle(img2, (x, y), params.line_thickness, (255, 0, 255), -1)

    _debug(visual=img2,
           filename=os.path.join(params.debug_outdir,
                                 str(params.device) + '_acute_vertices.png'))

    # Store into global measurements
    outputs.add_observation(sample=label,
                            variable='tip_coordinates',
                            trait='tip coordinates',
                            method='plantcv.plantcv.acute_vertex',
                            scale='none',
                            datatype=list,
                            value=acute_points,
                            label='none')

    return acute_points, img2
